import json

SILO_LENS_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "contract ISiloRepository",
        "name": "_siloRepo",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "DifferentArrayLength",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidRepository",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnsupportedLTVType",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroAssets",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_assetTotalDeposits",
        "type": "uint256"
      },
      {
        "internalType": "contract IShareToken",
        "name": "_shareToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "balanceOfUnderlying",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "borrowAPY",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "borrowShare",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "calcFee",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "calculateBorrowValue",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "calculateCollateralValue",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "collateralBalanceOfUnderlying",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "collateralOnlyDeposits",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "debtBalanceOfUnderlying",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "depositAPY",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_timestamp",
        "type": "uint256"
      }
    ],
    "name": "getBorrowAmount",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "getModel",
    "outputs": [
      {
        "internalType": "contract IInterestRateModel",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "getUserLTV",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "userLTV",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "getUserLiquidationThreshold",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "liquidationThreshold",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "getUserMaximumLTV",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "maximumLTV",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "getUtilization",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "hasPosition",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "inDebt",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "lensPing",
    "outputs": [
      {
        "internalType": "bytes4",
        "name": "",
        "type": "bytes4"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "liquidity",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "protocolFees",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "siloRepository",
    "outputs": [
      {
        "internalType": "contract ISiloRepository",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "totalBorrowAmount",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "totalBorrowAmountWithInterest",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_totalBorrowAmount",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "totalBorrowShare",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "totalDeposits",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISilo",
        "name": "_silo",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "totalDepositsWithInterest",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_totalDeposits",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')



import json

POOLREGISTRY_IMPLEMENTATION = json.loads("""
[
  {
    "inputs": [],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "calledContract",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "methodSignature",
        "type": "string"
      }
    ],
    "name": "Unauthorized",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroAddressNotAllowed",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "vTokenAddress",
        "type": "address"
      }
    ],
    "name": "MarketAdded",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "oldAccessControlManager",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "newAccessControlManager",
        "type": "address"
      }
    ],
    "name": "NewAccessControlManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferStarted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "string",
            "name": "category",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "logoURL",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "description",
            "type": "string"
          }
        ],
        "indexed": false,
        "internalType": "struct PoolRegistryInterface.VenusPoolMetaData",
        "name": "oldMetadata",
        "type": "tuple"
      },
      {
        "components": [
          {
            "internalType": "string",
            "name": "category",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "logoURL",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "description",
            "type": "string"
          }
        ],
        "indexed": false,
        "internalType": "struct PoolRegistryInterface.VenusPoolMetaData",
        "name": "newMetadata",
        "type": "tuple"
      }
    ],
    "name": "PoolMetadataUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "string",
        "name": "oldName",
        "type": "string"
      },
      {
        "indexed": false,
        "internalType": "string",
        "name": "newName",
        "type": "string"
      }
    ],
    "name": "PoolNameSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "string",
            "name": "name",
            "type": "string"
          },
          {
            "internalType": "address",
            "name": "creator",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "comptroller",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blockPosted",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "timestampPosted",
            "type": "uint256"
          }
        ],
        "indexed": false,
        "internalType": "struct PoolRegistryInterface.VenusPool",
        "name": "pool",
        "type": "tuple"
      }
    ],
    "name": "PoolRegistered",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "acceptOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "accessControlManager",
    "outputs": [
      {
        "internalType": "contract IAccessControlManagerV8",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "contract VToken",
            "name": "vToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactor",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "liquidationThreshold",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "initialSupply",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "vTokenReceiver",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "supplyCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          }
        ],
        "internalType": "struct PoolRegistry.AddMarketInput",
        "name": "input",
        "type": "tuple"
      }
    ],
    "name": "addMarket",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "string",
        "name": "name",
        "type": "string"
      },
      {
        "internalType": "contract Comptroller",
        "name": "comptroller",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "closeFactor",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "liquidationIncentive",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "minLiquidatableCollateral",
        "type": "uint256"
      }
    ],
    "name": "addPool",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "index",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAllPools",
    "outputs": [
      {
        "components": [
          {
            "internalType": "string",
            "name": "name",
            "type": "string"
          },
          {
            "internalType": "address",
            "name": "creator",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "comptroller",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blockPosted",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "timestampPosted",
            "type": "uint256"
          }
        ],
        "internalType": "struct PoolRegistryInterface.VenusPool[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      }
    ],
    "name": "getPoolByComptroller",
    "outputs": [
      {
        "components": [
          {
            "internalType": "string",
            "name": "name",
            "type": "string"
          },
          {
            "internalType": "address",
            "name": "creator",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "comptroller",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blockPosted",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "timestampPosted",
            "type": "uint256"
          }
        ],
        "internalType": "struct PoolRegistryInterface.VenusPool",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getPoolsSupportedByAsset",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getVTokenForAsset",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      }
    ],
    "name": "getVenusPoolMetadata",
    "outputs": [
      {
        "components": [
          {
            "internalType": "string",
            "name": "category",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "logoURL",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "description",
            "type": "string"
          }
        ],
        "internalType": "struct PoolRegistryInterface.VenusPoolMetaData",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "accessControlManager_",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "metadata",
    "outputs": [
      {
        "internalType": "string",
        "name": "category",
        "type": "string"
      },
      {
        "internalType": "string",
        "name": "logoURL",
        "type": "string"
      },
      {
        "internalType": "string",
        "name": "description",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pendingOwner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "accessControlManager_",
        "type": "address"
      }
    ],
    "name": "setAccessControlManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "name",
        "type": "string"
      }
    ],
    "name": "setPoolName",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "comptroller",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "string",
            "name": "category",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "logoURL",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "description",
            "type": "string"
          }
        ],
        "internalType": "struct PoolRegistryInterface.VenusPoolMetaData",
        "name": "metadata_",
        "type": "tuple"
      }
    ],
    "name": "updatePoolMetadata",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
""")
        


SPOOKYSWAP_MASTERCHEF_V2_ABI = [{
    "inputs": [{
        "internalType": "contract IMasterChef", "name": "_MASTER_CHEF",
        "type": "address"},
        {"internalType": "contract IERC20", "name": "_boo", "type": "address"},
        {"internalType": "uint256", "name": "_MASTER_PID", "type": "uint256"}],
    "stateMutability": "nonpayable", "type": "constructor"}, {
    "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "by", "type": "address"},
        {"indexed": True, "internalType": "address", "name": "to", "type": "address"}], "name": "AddAuth",
    "type": "event"}, {
    "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
        {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
        {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"},
        {"indexed": True, "internalType": "address", "name": "to", "type": "address"}], "name": "Deposit",
    "type": "event"}, {
    "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
        {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
        {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"},
        {"indexed": True, "internalType": "address", "name": "to", "type": "address"}], "name": "EmergencyWithdraw",
    "type": "event"}, {
    "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
        {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
        {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "Harvest",
    "type": "event"},
    {"anonymous": False, "inputs": [], "name": "LogInit", "type": "event"}, {
        "anonymous": False, "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "allocPoint", "type": "uint256"},
            {"indexed": True, "internalType": "contract IERC20", "name": "lpToken", "type": "address"},
            {"indexed": False, "internalType": "contract IRewarder", "name": "rewarder", "type": "address"},
            {"indexed": False, "internalType": "bool", "name": "update", "type": "bool"}], "name": "LogPoolAddition",
        "type": "event"}, {
        "anonymous": False, "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "allocPoint", "type": "uint256"},
            {"indexed": False, "internalType": "contract IRewarder", "name": "rewarder", "type": "address"},
            {"indexed": False, "internalType": "bool", "name": "overwrite", "type": "bool"},
            {"indexed": False, "internalType": "bool", "name": "update", "type": "bool"}], "name": "LogSetPool",
        "type": "event"}, {
        "anonymous": False, "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "lastRewardTime", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "lpSupply", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "accBooPerShare", "type": "uint256"}],
        "name": "LogUpdatePool", "type": "event"}, {
        "anonymous": False, "inputs": [
            {"indexed": True, "internalType": "address", "name": "by", "type": "address"},
            {"indexed": True, "internalType": "address", "name": "to", "type": "address"}], "name": "RevokeAuth",
        "type": "event"}, {
        "anonymous": False, "inputs": [
            {"indexed": True, "internalType": "address", "name": "by", "type": "address"},
            {"indexed": True, "internalType": "address", "name": "to", "type": "address"}], "name": "SetAdmin",
        "type": "event"}, {
        "anonymous": False, "inputs": [
            {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
            {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"},
            {"indexed": True, "internalType": "address", "name": "to", "type": "address"}], "name": "Withdraw",
        "type": "event"}, {
        "inputs": [], "name": "BOO",
        "outputs": [{"internalType": "contract IERC20", "name": "", "type": "address"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "MASTER_CHEF", "outputs": [
            {"internalType": "contract IMasterChef", "name": "", "type": "address"}], "stateMutability": "view",
        "type": "function"}, {
        "inputs": [], "name": "MASTER_PID",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "V1_HARVEST_QUERY_TIME",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint64", "name": "allocPoint", "type": "uint64"}, {
            "internalType": "contract IERC20", "name": "_lpToken", "type": "address"}, {
                       "internalType": "contract IRewarder", "name": "_rewarder",
                       "type": "address"},
                   {"internalType": "bool", "name": "update", "type": "bool"}],
        "name": "add", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
    {
        "inputs": [{"internalType": "address", "name": "_auth", "type": "address"}],
        "name": "addAuth", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [], "name": "admin",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "authorized",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "booPerSecond",
        "outputs": [{"internalType": "uint256", "name": "amount", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"},
                   {"internalType": "uint256", "name": "amount", "type": "uint256"},
                   {"internalType": "address", "name": "to", "type": "address"}],
        "name": "deposit", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"},
                   {"internalType": "uint256", "name": "amount", "type": "uint256"}],
        "name": "deposit", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"},
                   {"internalType": "address", "name": "to", "type": "address"}],
        "name": "emergencyWithdraw", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"}],
        "name": "emergencyWithdraw", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"}],
        "name": "getFarmData", "outputs": [{
            "components": [{
                "internalType": "uint128",
                "name": "accBooPerShare",
                "type": "uint128"}, {
                "internalType": "uint64",
                "name": "lastRewardTime",
                "type": "uint64"}, {
                "internalType": "uint64",
                "name": "allocPoint",
                "type": "uint64"}],
            "internalType": "struct MasterChefV2.PoolInfo",
            "name": "", "type": "tuple"}, {
            "internalType": "uint256", "name": "",
            "type": "uint256"}, {
            "internalType": "contract IRewarder",
            "name": "", "type": "address"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "harvestAll", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [], "name": "harvestFromMasterChef", "outputs": [],
        "stateMutability": "nonpayable", "type": "function"}, {
        "inputs": [{"internalType": "uint256[]", "name": "pids", "type": "uint256[]"}],
        "name": "harvestMultiple", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [
            {"internalType": "contract IERC20", "name": "dummyToken", "type": "address"}],
        "name": "init", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
    {
        "inputs": [{"internalType": "address", "name": "", "type": "address"}],
        "name": "isAuth", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "contract IERC20", "name": "", "type": "address"}],
        "name": "isLpToken",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "lastV1HarvestTimestamp",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "lpSupplies",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "lpToken",
        "outputs": [{"internalType": "contract IERC20", "name": "", "type": "address"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "massUpdateAllPools", "outputs": [],
        "stateMutability": "nonpayable", "type": "function"}, {
        "inputs": [{"internalType": "uint256[]", "name": "pids", "type": "uint256[]"}],
        "name": "massUpdatePools", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "bytes[]", "name": "data", "type": "bytes[]"}],
        "name": "multicall",
        "outputs": [{"internalType": "bytes[]", "name": "results", "type": "bytes[]"}],
        "stateMutability": "payable", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"},
                   {"internalType": "address", "name": "_user", "type": "address"}],
        "name": "pendingBOO",
        "outputs": [{"internalType": "uint256", "name": "pending", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "poolInfo", "outputs": [
            {"internalType": "uint128", "name": "accBooPerShare", "type": "uint128"},
            {"internalType": "uint64", "name": "lastRewardTime", "type": "uint64"},
            {"internalType": "uint64", "name": "allocPoint", "type": "uint64"}], "stateMutability": "view",
        "type": "function"}, {
        "inputs": [], "name": "poolInfoAmount",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "poolLength",
        "outputs": [{"internalType": "uint256", "name": "pools", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [], "name": "queryHarvestFromMasterChef", "outputs": [],
        "stateMutability": "nonpayable", "type": "function"}, {
        "inputs": [{"internalType": "address", "name": "_auth", "type": "address"}],
        "name": "revokeAuth", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "rewarder",
        "outputs": [{"internalType": "contract IRewarder", "name": "", "type": "address"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                   {"internalType": "uint256", "name": "value", "type": "uint256"},
                   {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                   {"internalType": "uint8", "name": "v", "type": "uint8"},
                   {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                   {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
        "name": "selfPermit", "outputs": [], "stateMutability": "payable",
        "type": "function"}, {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                   {"internalType": "uint256", "name": "nonce", "type": "uint256"},
                   {"internalType": "uint256", "name": "expiry", "type": "uint256"},
                   {"internalType": "uint8", "name": "v", "type": "uint8"},
                   {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                   {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
        "name": "selfPermitAllowed", "outputs": [], "stateMutability": "payable",
        "type": "function"}, {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                   {"internalType": "uint256", "name": "nonce", "type": "uint256"},
                   {"internalType": "uint256", "name": "expiry", "type": "uint256"},
                   {"internalType": "uint8", "name": "v", "type": "uint8"},
                   {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                   {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
        "name": "selfPermitAllowedIfNecessary", "outputs": [], "stateMutability": "payable",
        "type": "function"}, {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                   {"internalType": "uint256", "name": "value", "type": "uint256"},
                   {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                   {"internalType": "uint8", "name": "v", "type": "uint8"},
                   {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                   {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
        "name": "selfPermitIfNecessary", "outputs": [], "stateMutability": "payable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"},
                   {"internalType": "uint64", "name": "_allocPoint", "type": "uint64"}, {
                       "internalType": "contract IRewarder", "name": "_rewarder",
                       "type": "address"},
                   {"internalType": "bool", "name": "overwrite", "type": "bool"},
                   {"internalType": "bool", "name": "update", "type": "bool"}],
        "name": "set", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
    {
        "inputs": [{"internalType": "address", "name": "newAdmin", "type": "address"}],
        "name": "setAdmin", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256[]", "name": "_pid", "type": "uint256[]"},
                   {"internalType": "uint64[]", "name": "_allocPoint", "type": "uint64[]"},
                   {
                       "internalType": "contract IRewarder[]", "name": "_rewarders",
                       "type": "address[]"},
                   {"internalType": "bool[]", "name": "overwrite", "type": "bool[]"},
                   {"internalType": "bool", "name": "update", "type": "bool"}],
        "name": "setBatch", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256[]", "name": "_pid", "type": "uint256[]"},
                   {"internalType": "uint64[]", "name": "_allocPoint", "type": "uint64[]"},
                   {"internalType": "bool", "name": "update", "type": "bool"}],
        "name": "setBatchSlim", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "newTime", "type": "uint256"},
                   {"internalType": "bool", "name": "inDays", "type": "bool"}],
        "name": "setV1HarvestQueryTime", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "supportsPermits", "outputs": [
            {"internalType": "bytes32", "name": "domainSeparator", "type": "bytes32"}], "stateMutability": "view",
        "type": "function"}, {
        "inputs": [], "name": "totalAllocPoint",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"}],
        "name": "updatePool", "outputs": [{
            "components": [{
                "internalType": "uint128",
                "name": "accBooPerShare",
                "type": "uint128"}, {
                "internalType": "uint64",
                "name": "lastRewardTime",
                "type": "uint64"}, {
                "internalType": "uint64",
                "name": "allocPoint",
                "type": "uint64"}],
            "internalType": "struct MasterChefV2.PoolInfo",
            "name": "pool", "type": "tuple"}],
        "stateMutability": "nonpayable", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"},
                   {"internalType": "address", "name": "", "type": "address"}],
        "name": "userInfo",
        "outputs": [{"internalType": "uint256", "name": "amount", "type": "uint256"},
                    {"internalType": "uint256", "name": "rewardDebt", "type": "uint256"}],
        "stateMutability": "view", "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"},
                   {"internalType": "uint256", "name": "amount", "type": "uint256"},
                   {"internalType": "address", "name": "to", "type": "address"}],
        "name": "withdraw", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}, {
        "inputs": [{"internalType": "uint256", "name": "pid", "type": "uint256"},
                   {"internalType": "uint256", "name": "amount", "type": "uint256"}],
        "name": "withdraw", "outputs": [], "stateMutability": "nonpayable",
        "type": "function"}]



from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.vault_constant import Vault
from defi_services.services.vault.trava_vault_services import TravaVaultStateService


class VaultServices:
    # chain
    bsc = {
        Vault.trava_vault: TravaVaultStateService
    }

    ethereum = {
        Vault.trava_vault: TravaVaultStateService
    }

    fantom = {
        Vault.trava_vault: TravaVaultStateService
    }

    mapping = {
        Chain.bsc: bsc,
        Chain.ethereum: ethereum,
        Chain.fantom: fantom
    }



from defi_services.abis.dex.pancakeswap.factory_v3_abi import PANCAKESWAP_V3_FACTORY_ABI
from defi_services.abis.dex.pancakeswap.masterchef_v3_abi import PANCAKESWAP_MASTERCHEF_V3_ABI
from defi_services.abis.dex.pancakeswap.nft_token_abi import PANCAKE_V3_NON_FUNGIBLE_POSITION_TOKEN_ABI
from defi_services.abis.dex.pancakeswap.pancakeswap_masterchef_v2_abi import PANCAKESWAP_MASTERCHEF_V2_ABI
from defi_services.abis.dex.pancakeswap.pancakeswap_v2_factory_abi import PANCAKESWAP_V2_FACTORY_ABI
from defi_services.abis.dex.pancakeswap.v3_pool_abi import PANCAKESWAP_V3_POOL_ABI

PANCAKESWAP_V0_BSC_INFO = {
    "project_id": "pancakeswap",
    "chain_id": "0x1",
    "router_address": "0xeff92a263d31888d860bd50809a8d171709b7b1c",
    'master_chef_address': '0x73feaa1ee314f8c655e354234017be2193c9e24e',
    "reward_token": "0x0e09fabb73bd3ade0a17ecc321fd13a19e81ce82",
    'forked': 'uniswap-v2'

}

PANCAKESWAP_V2_BSC_INFO = {
    "project_id": "pancakeswap-amm",
    "chain_id": "0x38",
    'factory_address': '0xca143ce32fe78f1f7019d7d551a6402fc5350c73',
    'master_chef_address': '0xa5f8c5dbd5f286960b9d90548680ae5ebff07652',
    "reward_token": "0x0e09fabb73bd3ade0a17ecc321fd13a19e81ce82",
    'factory_abi': PANCAKESWAP_V2_FACTORY_ABI,
    'master_chef_abi': PANCAKESWAP_MASTERCHEF_V2_ABI,
    'forked': 'uniswap-v2',
    "router_address": "0x10ed43c718714eb63d5aa57b78b54704e256024e",
}

PANCAKESWAP_V3_BSC_INFO = {
    "project_id": "pancakeswap-v3",
    "chain_id": "0x38",
    'factory_address': '0x0bfbcf9fa4f9c56b0f40a671ad40e0805a091865',
    'factory_abi': PANCAKESWAP_V3_FACTORY_ABI,
    'master_chef_address': '0x556b9306565093c855aea9ae92a594704c2cd59e',
    'master_chef_abi': PANCAKESWAP_MASTERCHEF_V3_ABI,
    "NFT_manager_address": "0x46a15b0b27311cedf172ab29e4f4766fbe7f4364",
    "NFT_manager_abi": PANCAKE_V3_NON_FUNGIBLE_POSITION_TOKEN_ABI,
    'pool_abi': PANCAKESWAP_V3_POOL_ABI,
    'forked': 'uniswap-v3',
    "reward_token": "0x0e09fabb73bd3ade0a17ecc321fd13a19e81ce82",
    'router_address': "0x13f4ea83d0bd40e75c8222255bc855a974568dd4"

}



import json

LIDO_STONKS_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "agent_",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "manager_",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "tokenFrom_",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "tokenTo_",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "amountConverter_",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "orderSample_",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "orderDurationInSeconds_",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "marginInBasisPoints_",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "priceToleranceInBasisPoints_",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "agent_",
        "type": "address"
      }
    ],
    "name": "InvalidAgentAddress",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "InvalidAmount",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "amountConverter",
        "type": "address"
      }
    ],
    "name": "InvalidAmountConverterAddress",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      }
    ],
    "name": "InvalidManagerAddress",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "min",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "max",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "received",
        "type": "uint256"
      }
    ],
    "name": "InvalidOrderDuration",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "orderSample",
        "type": "address"
      }
    ],
    "name": "InvalidOrderSampleAddress",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "tokenFrom",
        "type": "address"
      }
    ],
    "name": "InvalidTokenFromAddress",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "tokenTo",
        "type": "address"
      }
    ],
    "name": "InvalidTokenToAddress",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "limit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "received",
        "type": "uint256"
      }
    ],
    "name": "MarginOverflowsAllowedLimit",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "min",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "received",
        "type": "uint256"
      }
    ],
    "name": "MinimumPossibleBalanceNotMet",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "NotAgent",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "NotAgentOrManager",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "limit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "received",
        "type": "uint256"
      }
    ],
    "name": "PriceToleranceOverflowsAllowedLimit",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TokensCannotBeSame",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "agent",
        "type": "address"
      }
    ],
    "name": "AgentSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "amountConverter",
        "type": "address"
      }
    ],
    "name": "AmountConverterSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ERC1155Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ERC20Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      }
    ],
    "name": "ERC721Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "EtherRecovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "manager",
        "type": "address"
      }
    ],
    "name": "ManagerSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "marginInBasisPoints",
        "type": "uint256"
      }
    ],
    "name": "MarginInBasisPointsSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "orderContract",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "minBuyAmount",
        "type": "uint256"
      }
    ],
    "name": "OrderContractCreated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "orderDurationInSeconds",
        "type": "uint256"
      }
    ],
    "name": "OrderDurationInSecondsSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "orderSample",
        "type": "address"
      }
    ],
    "name": "OrderSampleSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "priceToleranceInBasisPoints",
        "type": "uint256"
      }
    ],
    "name": "PriceToleranceInBasisPointsSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "tokenFrom",
        "type": "address"
      }
    ],
    "name": "TokenFromSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "tokenTo",
        "type": "address"
      }
    ],
    "name": "TokenToSet",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "AGENT",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "AMOUNT_CONVERTER",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "MARGIN_IN_BASIS_POINTS",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "ORDER_DURATION_IN_SECONDS",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "ORDER_SAMPLE",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "PRICE_TOLERANCE_IN_BASIS_POINTS",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "TOKEN_FROM",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "TOKEN_TO",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "amount_",
        "type": "uint256"
      }
    ],
    "name": "estimateTradeOutput",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "estimatedTradeOutput",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "estimateTradeOutputFromCurrentBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getOrderParameters",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getPriceTolerance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "manager",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "minBuyAmount_",
        "type": "uint256"
      }
    ],
    "name": "placeOrder",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "token_",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "tokenId_",
        "type": "uint256"
      }
    ],
    "name": "recoverERC1155",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "token_",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount_",
        "type": "uint256"
      }
    ],
    "name": "recoverERC20",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "token_",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "tokenId_",
        "type": "uint256"
      }
    ],
    "name": "recoverERC721",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "recoverEther",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "manager_",
        "type": "address"
      }
    ],
    "name": "setManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


import csv
import json
import time
from collections import defaultdict
from typing import Tuple, List, Union, Any, Optional, Dict

from query_state_lib.base.utils.decoder import decode_eth_call_data
from query_state_lib.client.client_querier import ClientQuerier
from web3 import Web3, contract

from defi_services.abis.dex.pancakeswap.pancakeswap_lp_token_abi import LP_TOKEN_ABI
from defi_services.abis.multicall_v3_abi import MULTICALL_V3_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.network_constants import MulticallContract, Chains, Networks
from defi_services.services.multicall.batch_queries_service import add_rpc_call, decode_data_response, \
    decode_data_response_ignore_error
from defi_services.utils.dict_utils import all_equal
from defi_services.utils.logger_utils import get_logger

logger = get_logger('Multicall V2')

w3 = Web3()


def get_fn_abi(abi, fn_name):
    for abi_ in abi:
        if abi_.get('type') == 'function' and abi_.get('name') == fn_name:
            return abi_

    return None


def _unpack_aggregate_outputs(outputs: Any) -> Tuple[Tuple[Union[None, bool], bytes], ...]:
    return tuple((success, output) for success, output in outputs)


class W3Multicall:
    """
    Interface for multicall3.sol contract
    """

    class Call:

        def __init__(self, address, abi, fn_name, fn_paras=None, block_number: Union[int, str] = 'latest', key=None):
            if key:
                self.id = key
            elif fn_paras is not None:
                self.id = f"{fn_name}_{address}_{fn_paras}_{block_number}".lower()
            else:
                self.id = f"{fn_name}_{address}_{block_number}".lower()

            self.address = Web3.to_checksum_address(address)
            self.fn_name = fn_name
            self.fn_abi = get_fn_abi(abi=abi, fn_name=fn_name)
            self.block_number = block_number

            args = []
            if fn_paras is not None:
                if type(fn_paras) is list:
                    for item in fn_paras:
                        if Web3.is_address(item):
                            args.append(Web3.to_checksum_address(item))
                        else:
                            args.append(item)
                else:
                    if Web3.is_address(fn_paras):
                        fn_paras = Web3.to_checksum_address(fn_paras)
                    args = [fn_paras]

            c = contract.Contract
            c.w3 = w3
            c.abi = abi

            self.data = c.encodeABI(fn_name=fn_name, args=args)

    def __init__(self, web3, address='0xcA11bde05977b3631167028862bE2a173976CA11', calls: Dict[str, 'W3Multicall.Call'] = None, require_success: bool = False):
        """
        :param web3: Web3 instance
        :param address: (optional) address of the multicall3.sol contract
        :param calls: (optional) list of W3Multicall.Call to perform
        """
        self.web3 = web3
        self.address = address
        self.calls: Dict[str, 'W3Multicall.Call'] = {} if calls is None else calls.copy()

        self.require_success = require_success

    def add(self, call: 'W3Multicall.Call'):
        self.calls[call.id] = call

    def get_params(self, calls: Optional[Dict[str, Call]] = None) -> List[Union[bool, List[List[Any]]]]:
        args = self._get_args(calls=calls)
        return args

    def decode(self, aggregated, calls: Optional[Dict[str, Call]] = None, ignore_error=True):
        if calls is None:
            calls = self.calls

        unpacked = _unpack_aggregate_outputs(aggregated)

        outputs = {}
        for (call_id, call), (success, output) in zip(calls.items(), unpacked):
            if not success:
                logger.warning(f'Fail to query {call_id}')
                continue

            try:
                call_output = '0x' + output.hex()
                decoded_data = decode_eth_call_data([call.fn_abi], call.fn_name, call_output)
            except OverflowError:
                bytes_array = bytearray.fromhex(output.hex())
                bytes32 = '0x' + bytes_array.hex().rstrip("0")
                if len(bytes32) % 2 != 0:
                    bytes32 += '0'
                decoded_data = decode_eth_call_data([call.fn_abi], call.fn_name, bytes32)
            except Exception as ex:
                logger.error(f"An exception when decode data from provider: {ex}")
                if ignore_error:
                    continue

                raise ex

            if len(decoded_data) == 1:
                decoded_data = decoded_data[0]

            outputs[call_id] = decoded_data

        return outputs

    def _get_args(self, calls: Optional[Dict[str, Call]] = None) -> List[Union[bool, List[List[Any]]]]:
        if calls is None:
            calls = self.calls

        block_numbers = [call.block_number for call in calls.values()]
        if not all_equal(block_numbers):
            raise ValueError('All calls must be queried at the same block number')

        return [self.require_success, [[call.address, call.data] for call in calls.values()]]

    def batch_calls_iterator(self, batch_size=2000):
        calls_by_block_number = defaultdict(lambda: {})
        for call_id, call in self.calls.items():
            block_number = call.block_number
            calls_by_block_number[block_number][call_id] = call

        for block_number, calls in calls_by_block_number.items():
            batch = {}

            for call_id, call in calls.items():
                batch[call_id] = call
                if len(batch) >= batch_size:
                    yield block_number, batch
                    batch = {}

            if len(batch) > 0:
                yield block_number, batch


def add_rpc_multicall(w3_multicall: W3Multicall, list_rpc_call=None, list_call_id=None, batch_size=2000):
    batch_idx = 0
    for block_number, batch_calls in w3_multicall.batch_calls_iterator(batch_size=batch_size):
        inputs = w3_multicall.get_params(calls=batch_calls)
        add_rpc_call(
            fn_name="tryAggregate", fn_paras=inputs, block_number=block_number,
            abi=MULTICALL_V3_ABI, contract_address=w3_multicall.address,
            list_call_id=list_call_id, list_rpc_call=list_rpc_call, call_id=f'{batch_idx}_{block_number}'
        )
        batch_idx += 1


def decode_multical_response(w3_multicall: W3Multicall, data_responses, list_call_id, ignore_error=True, batch_size=2000):
    if ignore_error:
        decoded_data = decode_data_response_ignore_error(data_responses=data_responses, list_call_id=list_call_id)
    else:
        decoded_data = decode_data_response(data_responses=data_responses, list_call_id=list_call_id)

    batch_idx = 0
    results = {}
    for block_number, batch_calls in w3_multicall.batch_calls_iterator(batch_size=batch_size):
        multicall_data = decoded_data.get(f'{batch_idx}_{block_number}')
        decode_multicall_data = w3_multicall.decode(multicall_data, calls=batch_calls, ignore_error=ignore_error)
        results.update(decode_multicall_data)

        batch_idx += 1

    return results


def _test_multicall(_w3: Web3, multicall_contract, wallets, tokens):
    start_time = time.time()
    encode_time = 0

    rpc_call_id = {}

    n_calls = 0
    data = {}
    for wallet in wallets:
        start_encode_time = time.time()
        data[wallet] = {}

        w3_multicall = W3Multicall(_w3, address=MulticallContract.get_multicall_contract(Chains.arbitrum))
        for token in tokens:
            # if token['address'] == '0x0000000000000000000000000000000000000000':
            #     continue

            w3_multicall.add(W3Multicall.Call(
                Web3.to_checksum_address(token['address']),  # contract address
                ERC20_ABI,  # method signature to call
                'balanceOf',
                wallet
            ))

            decimals_call = W3Multicall.Call(
                Web3.to_checksum_address(token['address']),  # contract address
                ERC20_ABI,  # method signature to call
                'decimals'
            )
            if decimals_call not in rpc_call_id:
                w3_multicall.add(decimals_call)
                rpc_call_id[decimals_call] = None

        encode_time += time.time() - start_encode_time
        n_calls += len(w3_multicall.calls)

        # For single call
        inputs = w3_multicall.get_params()
        response = multicall_contract.functions.tryAggregate(*inputs).call()
        results = w3_multicall.decode(response)

        for token in tokens:
            # if token['address'] == '0x0000000000000000000000000000000000000000':
            #     continue

            decimals_call_id = f'decimals_{token["address"]}_latest'.lower()
            balance_call_id = f'balanceOf_{token["address"]}_{wallet}_latest'.lower()

            decimals = results.get(decimals_call_id)
            if decimals is None:
                decimals = rpc_call_id.get(decimals_call_id) or 18
            else:
                rpc_call_id[decimals_call_id] = decimals

            balance = results.get(balance_call_id)
            if balance is None:
                balance = rpc_call_id.get(balance_call_id) or 0
            else:
                rpc_call_id[balance_call_id] = balance
            balance /= 10 ** decimals

            data[wallet][token['address']] = {
                'name': token['name'],
                'symbol': token['symbol'],
                'balance': balance,
                'decimals': decimals
            }

    with open('results_multicall.json', 'w') as f:
        json.dump(data, f, indent=2)
    duration = time.time() - start_time

    print(f'There are {n_calls} calls')
    print(f'Encode took {round(encode_time, 3)}s')
    print(f'Done after {round(duration, 3)}s')

    return duration, encode_time, n_calls


def _test_state_querier(client_querier: ClientQuerier, wallets, tokens):
    start_time = time.time()

    list_call_id = []
    list_rpc_call = []

    for wallet in wallets:
        for token in tokens:
            if token['address'] == '0x0000000000000000000000000000000000000000':
                continue

            add_rpc_call(
                abi=ERC20_ABI, contract_address=Web3.to_checksum_address(token['address']),
                fn_name="balanceOf", fn_paras=wallet, block_number='latest',
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )
            add_rpc_call(
                abi=ERC20_ABI, contract_address=Web3.to_checksum_address(token['address']),
                fn_name="decimals", block_number='latest',
                list_call_id=list_call_id, list_rpc_call=list_rpc_call
            )

    encode_time = time.time() - start_time
    print(f'There are {len(list_rpc_call)} calls')
    print(f'Encode took {round(encode_time, 3)}s')

    responses = client_querier.sent_batch_to_provider(list_rpc_call, batch_size=2000)
    decoded_data = decode_data_response_ignore_error(data_responses=responses, list_call_id=list_call_id)

    data = {}
    for wallet in wallets:
        data[wallet] = {}
        for token in tokens:
            if token['address'] == '0x0000000000000000000000000000000000000000':
                continue

            decimals = decoded_data.get(f'decimals_{token["address"]}_latest'.lower())
            balance = decoded_data.get(f'balanceOf_{token["address"]}_{wallet}_latest'.lower()) / 10 ** decimals

            data[wallet][token['address']] = {
                'name': token['name'],
                'symbol': token['symbol'],
                'balance': balance,
                'decimals': decimals
            }

    with open('results_querier.json', 'w') as f:
        json.dump(data, f, indent=2)
    duration = time.time() - start_time
    print(f'Done after {round(duration, 3)}s')

    return duration, encode_time, len(list_rpc_call)


def _test_multicall_with_multiprocessing(_w3: Web3, client_querier: ClientQuerier, wallets, tokens):
    start_time = time.time()

    w3_multicall = W3Multicall(_w3, address=MulticallContract.get_multicall_contract(Chains.arbitrum))
    for wallet in wallets:
        for token in tokens:
            # if token['address'] == '0x0000000000000000000000000000000000000000':
            #     continue

            w3_multicall.add(W3Multicall.Call(
                Web3.to_checksum_address(token['address']),  # contract address
                ERC20_ABI,  # method signature to call
                'balanceOf',
                wallet
            ))

            w3_multicall.add(W3Multicall.Call(
                Web3.to_checksum_address(token['address']),  # contract address
                ERC20_ABI,  # method signature to call
                'decimals'
            ))

            w3_multicall.add(W3Multicall.Call(
                Web3.to_checksum_address(token['address']),  # contract address
                LP_TOKEN_ABI,  # method signature to call
                'token0'
            ))

    list_call_id, list_rpc_call = [], []
    add_rpc_multicall(w3_multicall, list_rpc_call=list_rpc_call, list_call_id=list_call_id)

    encode_time = time.time() - start_time

    responses = client_querier.sent_batch_to_provider(list_rpc_call, batch_size=1)
    decoded_data = decode_multical_response(
        w3_multicall=w3_multicall, data_responses=responses,
        list_call_id=list_call_id, ignore_error=True
    )

    data = {}
    for wallet in wallets:
        data[wallet] = {}
        for token in tokens:
            # if token['address'] == '0x0000000000000000000000000000000000000000':
            #     continue

            decimals_call_id = f'decimals_{token["address"]}_latest'.lower()
            balance_call_id = f'balanceOf_{token["address"]}_{wallet}_latest'.lower()

            decimals = decoded_data.get(decimals_call_id) or 18
            balance = decoded_data.get(balance_call_id, 0) / 10 ** decimals

            data[wallet][token['address']] = {
                'name': token['name'],
                'symbol': token['symbol'],
                'balance': balance,
                'decimals': decimals
            }

    with open('results_multicall_with_multiprocessing.json', 'w') as f:
        json.dump(data, f, indent=2)
    duration = time.time() - start_time

    print(f'There are {len(w3_multicall.calls)} calls')
    print(f'Encode took {round(encode_time, 3)}s')
    print(f'Done after {round(duration, 3)}s')

    return duration, encode_time, len(w3_multicall.calls)


def query_multi_contracts(n_times=20, wallets_batch_size=5, wallets_distribute_similar=True):
    with open('../../../.data/wallets.json') as f:
        wallets = json.load(f)
    wallets = ['0x6FfA563915CB3186b9Dd206D0E08bdeDcd2EA2ED']

    with open('../../../.data/tokens.json') as f:
        tokens_ = json.load(f)
    tokens_ = [
        {
            "address": "0xE2035f04040A135c4dA2f96AcA742143c57c79F9",
            "name": "UXUY",
            "symbol": "uxuy"
        },
        {
            "address": "0xfd086bc7cd5c481dcc9c85ebe478a1c0b69fcbb9",
            "name": "Tether",
            "symbol": "usdt"
        },
        {
            "address": "0x0000000000000000000000000000000000000000",
            "name": "ETH",
            "symbol": "ETH"
        }
    ]

    print(f'There are maximum {len(wallets)} wallets')
    print(f'There are {len(tokens_)} tokens')

    if wallets_distribute_similar:
        wallets_batch = [wallets[:300], wallets[:300], wallets[:300]]
    else:
        wallets_batch = [wallets[:300], wallets[300:600], wallets[600:900]]

    # Methodology 1: Query State Lib
    start_time = time.time()
    client_querier = ClientQuerier(provider_url=Networks.providers['arbitrum'])
    starting_time_1 = time.time() - start_time
    logger.info(f'Starting state querier: {round(starting_time_1, 3)}s')

    _multiple(
        n_times=n_times, starting_time=starting_time_1, filename='query_state_lib',
        wallets=wallets_batch[0], wallets_batch_size=wallets_batch_size, tokens=tokens_, func=_test_state_querier,
        client_querier=client_querier
    )

    # Methodology 2: Multicall
    start_time = time.time()
    _w3 = Web3(Web3.HTTPProvider(Networks.providers['arbitrum']))
    starting_time_2_1 = time.time() - start_time
    multicall_address = MulticallContract.get_multicall_contract(Chains.arbitrum)
    multicall_contract = _w3.eth.contract(Web3.to_checksum_address(multicall_address), abi=MULTICALL_V3_ABI)
    starting_time_2 = time.time() - start_time
    logger.info(f'Starting multicall: {round(starting_time_2, 3)}s')

    _multiple(
        n_times=n_times, starting_time=starting_time_2, filename='multicall',
        wallets=wallets_batch[1], wallets_batch_size=wallets_batch_size, tokens=tokens_, func=_test_multicall,
        _w3=_w3, multicall_contract=multicall_contract
    )

    # Methodology 3: Combine Multicall with multiprocessing
    starting_time_3 = starting_time_1 + starting_time_2_1
    logger.info(f'Starting combined: {round(starting_time_3, 3)}s')

    _multiple(
        n_times=n_times, starting_time=starting_time_3, filename='combined',
        wallets=wallets_batch[2], wallets_batch_size=wallets_batch_size, tokens=tokens_,
        func=_test_multicall_with_multiprocessing, _w3=_w3, client_querier=client_querier
    )


def _multiple(n_times, starting_time, filename, wallets, wallets_batch_size, tokens, func, *args, **kwargs):
    data = [['Total Time', 'Encode Time']]
    overview = {
        'avg_total_time': 0, 'avg_encode_time': 0,
        'times': 0, 'errors': 0,
        'queries': 0, 'start_time': starting_time
    }
    for i in range(n_times):
        idx = i * wallets_batch_size
        sub_wallets = wallets[idx:idx + wallets_batch_size]
        try:
            total_time, encode_time, n_calls = func(wallets=sub_wallets, tokens=tokens, *args, **kwargs)

            overview['queries'] = n_calls
            data.append([total_time, encode_time])
            overview['times'] += 1
        except Exception as ex:
            logger.exception(ex)
            overview['errors'] += 1
        finally:
            time.sleep(13)

    with open(f'results_{filename}.csv', 'w') as f:
        writer = csv.writer(f)
        writer.writerows(data)

    overview['avg_total_time'] = sum([t[0] for t in data[1:]]) / overview['times']
    overview['avg_encode_time'] = sum([t[1] for t in data[1:]]) / overview['times']
    with open(f'overview_results_{filename}.json', 'w') as f:
        json.dump(overview, f, indent=2)


if __name__ == '__main__':
    query_multi_contracts(n_times=1, wallets_batch_size=1, wallets_distribute_similar=True)



from defi_services.abis.dex.quickswap.factory_v3_abi import QUICKSWAP_V3_FACTORY_ABI
from defi_services.abis.dex.quickswap.nft_token_manager import QUICKSWAP_V3_NON_FUNGIBLE_POSITION_TOKEN_ABI
from defi_services.abis.dex.quickswap.pool_v3_abi import QUICKSWAP_V3_POOL_ABI

QUICKSWAP_POLYGON_V2_INFO = {
    'factory_address': '0x5757371414417b8c6caad45baef941abc7d3ab32',
    'forked': 'uniswap-v2'

}

QUICKSWAP_POLYGON_V3_INFO = {
    'factory_address': '0x411b0facc3489691f28ad58c47006af5e3ab3a28',
    'factory_abi': QUICKSWAP_V3_FACTORY_ABI,
    "NFT_manager_address": "0x8ef88e4c7cfbbac1c163f7eddd4b578792201de6",
    "NFT_manager_abi": QUICKSWAP_V3_NON_FUNGIBLE_POSITION_TOKEN_ABI,
    'pool_abi':QUICKSWAP_V3_POOL_ABI,
    'finite_farming': '0x9923f42a02a82da63ee0dbbc5f8e311e3dd8a1f8',
    'infinite_farming': '0x8a26436e41d0b5fc4c6ed36c1976fafbe173444e',
    'farming_center': '0x7f281a8cdf66ef5e9db8434ec6d97acc1bc01e78',
    'forked': 'uniswap-v3'

}



import json

LIDO_BALANCER_WSETH_ETH_POOL_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "contract IVault",
            "name": "vault",
            "type": "address"
          },
          {
            "internalType": "string",
            "name": "name",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "symbol",
            "type": "string"
          },
          {
            "internalType": "contract IERC20[]",
            "name": "tokens",
            "type": "address[]"
          },
          {
            "internalType": "contract IRateProvider[]",
            "name": "rateProviders",
            "type": "address[]"
          },
          {
            "internalType": "uint256[]",
            "name": "priceRateCacheDuration",
            "type": "uint256[]"
          },
          {
            "internalType": "uint256",
            "name": "amplificationParameter",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "swapFeePercentage",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "pauseWindowDuration",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "bufferPeriodDuration",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "oracleEnabled",
            "type": "bool"
          },
          {
            "internalType": "address",
            "name": "owner",
            "type": "address"
          }
        ],
        "internalType": "struct MetaStablePool.NewPoolParams",
        "name": "params",
        "type": "tuple"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "startValue",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "endValue",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "startTime",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "endTime",
        "type": "uint256"
      }
    ],
    "name": "AmpUpdateStarted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "currentValue",
        "type": "uint256"
      }
    ],
    "name": "AmpUpdateStopped",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "value",
        "type": "uint256"
      }
    ],
    "name": "Approval",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bool",
        "name": "enabled",
        "type": "bool"
      }
    ],
    "name": "OracleEnabledChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bool",
        "name": "paused",
        "type": "bool"
      }
    ],
    "name": "PausedStateChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract IERC20",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rate",
        "type": "uint256"
      }
    ],
    "name": "PriceRateCacheUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract IERC20",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "contract IRateProvider",
        "name": "provider",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "cacheDuration",
        "type": "uint256"
      }
    ],
    "name": "PriceRateProviderSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "swapFeePercentage",
        "type": "uint256"
      }
    ],
    "name": "SwapFeePercentageChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "value",
        "type": "uint256"
      }
    ],
    "name": "Transfer",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "DOMAIN_SEPARATOR",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      }
    ],
    "name": "allowance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "approve",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "balanceOf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "decimals",
    "outputs": [
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "decreaseAllowance",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "enableOracle",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes4",
        "name": "selector",
        "type": "bytes4"
      }
    ],
    "name": "getActionId",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAmplificationParameter",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "value",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "isUpdating",
        "type": "bool"
      },
      {
        "internalType": "uint256",
        "name": "precision",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAuthorizer",
    "outputs": [
      {
        "internalType": "contract IAuthorizer",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getLargestSafeQueryWindow",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getLastInvariant",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "lastInvariant",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "lastInvariantAmp",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "enum IPriceOracle.Variable",
        "name": "variable",
        "type": "uint8"
      }
    ],
    "name": "getLatest",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getOracleMiscData",
    "outputs": [
      {
        "internalType": "int256",
        "name": "logInvariant",
        "type": "int256"
      },
      {
        "internalType": "int256",
        "name": "logTotalSupply",
        "type": "int256"
      },
      {
        "internalType": "uint256",
        "name": "oracleSampleCreationTimestamp",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "oracleIndex",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "oracleEnabled",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getOwner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "enum IPriceOracle.Variable",
            "name": "variable",
            "type": "uint8"
          },
          {
            "internalType": "uint256",
            "name": "ago",
            "type": "uint256"
          }
        ],
        "internalType": "struct IPriceOracle.OracleAccumulatorQuery[]",
        "name": "queries",
        "type": "tuple[]"
      }
    ],
    "name": "getPastAccumulators",
    "outputs": [
      {
        "internalType": "int256[]",
        "name": "results",
        "type": "int256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getPausedState",
    "outputs": [
      {
        "internalType": "bool",
        "name": "paused",
        "type": "bool"
      },
      {
        "internalType": "uint256",
        "name": "pauseWindowEndTime",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "bufferPeriodEndTime",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getPoolId",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IERC20",
        "name": "token",
        "type": "address"
      }
    ],
    "name": "getPriceRateCache",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "rate",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "duration",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "expires",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getRate",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getRateProviders",
    "outputs": [
      {
        "internalType": "contract IRateProvider[]",
        "name": "providers",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "index",
        "type": "uint256"
      }
    ],
    "name": "getSample",
    "outputs": [
      {
        "internalType": "int256",
        "name": "logPairPrice",
        "type": "int256"
      },
      {
        "internalType": "int256",
        "name": "accLogPairPrice",
        "type": "int256"
      },
      {
        "internalType": "int256",
        "name": "logBptPrice",
        "type": "int256"
      },
      {
        "internalType": "int256",
        "name": "accLogBptPrice",
        "type": "int256"
      },
      {
        "internalType": "int256",
        "name": "logInvariant",
        "type": "int256"
      },
      {
        "internalType": "int256",
        "name": "accLogInvariant",
        "type": "int256"
      },
      {
        "internalType": "uint256",
        "name": "timestamp",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getScalingFactors",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getSwapFeePercentage",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "enum IPriceOracle.Variable",
            "name": "variable",
            "type": "uint8"
          },
          {
            "internalType": "uint256",
            "name": "secs",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "ago",
            "type": "uint256"
          }
        ],
        "internalType": "struct IPriceOracle.OracleAverageQuery[]",
        "name": "queries",
        "type": "tuple[]"
      }
    ],
    "name": "getTimeWeightedAverage",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "results",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getTotalSamples",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getVault",
    "outputs": [
      {
        "internalType": "contract IVault",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "addedValue",
        "type": "uint256"
      }
    ],
    "name": "increaseAllowance",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "name",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      }
    ],
    "name": "nonces",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "poolId",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "balances",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256",
        "name": "lastChangeBlock",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "protocolSwapFeePercentage",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "userData",
        "type": "bytes"
      }
    ],
    "name": "onExitPool",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "amountsOut",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256[]",
        "name": "dueProtocolFeeAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "poolId",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "balances",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256",
        "name": "lastChangeBlock",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "protocolSwapFeePercentage",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "userData",
        "type": "bytes"
      }
    ],
    "name": "onJoinPool",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "amountsIn",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256[]",
        "name": "dueProtocolFeeAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "enum IVault.SwapKind",
            "name": "kind",
            "type": "uint8"
          },
          {
            "internalType": "contract IERC20",
            "name": "tokenIn",
            "type": "address"
          },
          {
            "internalType": "contract IERC20",
            "name": "tokenOut",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "amount",
            "type": "uint256"
          },
          {
            "internalType": "bytes32",
            "name": "poolId",
            "type": "bytes32"
          },
          {
            "internalType": "uint256",
            "name": "lastChangeBlock",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "from",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "to",
            "type": "address"
          },
          {
            "internalType": "bytes",
            "name": "userData",
            "type": "bytes"
          }
        ],
        "internalType": "struct IPoolSwapStructs.SwapRequest",
        "name": "request",
        "type": "tuple"
      },
      {
        "internalType": "uint256[]",
        "name": "balances",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256",
        "name": "indexIn",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "indexOut",
        "type": "uint256"
      }
    ],
    "name": "onSwap",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "enum IVault.SwapKind",
            "name": "kind",
            "type": "uint8"
          },
          {
            "internalType": "contract IERC20",
            "name": "tokenIn",
            "type": "address"
          },
          {
            "internalType": "contract IERC20",
            "name": "tokenOut",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "amount",
            "type": "uint256"
          },
          {
            "internalType": "bytes32",
            "name": "poolId",
            "type": "bytes32"
          },
          {
            "internalType": "uint256",
            "name": "lastChangeBlock",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "from",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "to",
            "type": "address"
          },
          {
            "internalType": "bytes",
            "name": "userData",
            "type": "bytes"
          }
        ],
        "internalType": "struct IPoolSwapStructs.SwapRequest",
        "name": "request",
        "type": "tuple"
      },
      {
        "internalType": "uint256",
        "name": "balanceTokenIn",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "balanceTokenOut",
        "type": "uint256"
      }
    ],
    "name": "onSwap",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "spender",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "value",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "deadline",
        "type": "uint256"
      },
      {
        "internalType": "uint8",
        "name": "v",
        "type": "uint8"
      },
      {
        "internalType": "bytes32",
        "name": "r",
        "type": "bytes32"
      },
      {
        "internalType": "bytes32",
        "name": "s",
        "type": "bytes32"
      }
    ],
    "name": "permit",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "poolId",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "balances",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256",
        "name": "lastChangeBlock",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "protocolSwapFeePercentage",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "userData",
        "type": "bytes"
      }
    ],
    "name": "queryExit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "bptIn",
        "type": "uint256"
      },
      {
        "internalType": "uint256[]",
        "name": "amountsOut",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "poolId",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "balances",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256",
        "name": "lastChangeBlock",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "protocolSwapFeePercentage",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "userData",
        "type": "bytes"
      }
    ],
    "name": "queryJoin",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "bptOut",
        "type": "uint256"
      },
      {
        "internalType": "uint256[]",
        "name": "amountsIn",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IERC20",
        "name": "token",
        "type": "address"
      },
      {
        "internalType": "bytes",
        "name": "poolConfig",
        "type": "bytes"
      }
    ],
    "name": "setAssetManagerPoolConfig",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bool",
        "name": "paused",
        "type": "bool"
      }
    ],
    "name": "setPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IERC20",
        "name": "token",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "duration",
        "type": "uint256"
      }
    ],
    "name": "setPriceRateCacheDuration",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "swapFeePercentage",
        "type": "uint256"
      }
    ],
    "name": "setSwapFeePercentage",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "rawEndValue",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "endTime",
        "type": "uint256"
      }
    ],
    "name": "startAmplificationParameterUpdate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "stopAmplificationParameterUpdate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "symbol",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalSupply",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "transfer",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "transferFrom",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IERC20",
        "name": "token",
        "type": "address"
      }
    ],
    "name": "updatePriceRateCache",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')



import json

COMPTROLLERIMPL = json.loads("""
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "poolRegistry_",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "market",
        "type": "address"
      },
      {
        "internalType": "enum ComptrollerStorage.Action",
        "name": "action",
        "type": "uint8"
      }
    ],
    "name": "ActionPaused",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "market",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "cap",
        "type": "uint256"
      }
    ],
    "name": "BorrowCapExceeded",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "expectedLessThanOrEqualTo",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "actual",
        "type": "uint256"
      }
    ],
    "name": "CollateralExceedsThreshold",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ComptrollerMismatch",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "collateralToSeize",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "availableCollateral",
        "type": "uint256"
      }
    ],
    "name": "InsufficientCollateral",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InsufficientLiquidity",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InsufficientShortfall",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidCollateralFactor",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidLiquidationThreshold",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "market",
        "type": "address"
      }
    ],
    "name": "MarketAlreadyListed",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "MarketNotCollateral",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "market",
        "type": "address"
      }
    ],
    "name": "MarketNotListed",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "loopsLimit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "requiredLoops",
        "type": "uint256"
      }
    ],
    "name": "MaxLoopsLimitExceeded",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "expectedGreaterThan",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "actual",
        "type": "uint256"
      }
    ],
    "name": "MinimalCollateralViolated",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NonzeroBorrowBalance",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "PriceError",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "SnapshotError",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "market",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "cap",
        "type": "uint256"
      }
    ],
    "name": "SupplyCapExceeded",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TooMuchRepay",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "calledContract",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "methodSignature",
        "type": "string"
      }
    ],
    "name": "Unauthorized",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "expectedSender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "actualSender",
        "type": "address"
      }
    ],
    "name": "UnexpectedSender",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroAddressNotAllowed",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "enum ComptrollerStorage.Action",
        "name": "action",
        "type": "uint8"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "pauseState",
        "type": "bool"
      }
    ],
    "name": "ActionPausedMarket",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "MarketEntered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "MarketExited",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "MarketSupported",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldMaxLoopsLimit",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newmaxLoopsLimit",
        "type": "uint256"
      }
    ],
    "name": "MaxLoopsLimitUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "oldAccessControlManager",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "newAccessControlManager",
        "type": "address"
      }
    ],
    "name": "NewAccessControlManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newBorrowCap",
        "type": "uint256"
      }
    ],
    "name": "NewBorrowCap",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldCloseFactorMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newCloseFactorMantissa",
        "type": "uint256"
      }
    ],
    "name": "NewCloseFactor",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldCollateralFactorMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newCollateralFactorMantissa",
        "type": "uint256"
      }
    ],
    "name": "NewCollateralFactor",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldLiquidationIncentiveMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newLiquidationIncentiveMantissa",
        "type": "uint256"
      }
    ],
    "name": "NewLiquidationIncentive",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldLiquidationThresholdMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newLiquidationThresholdMantissa",
        "type": "uint256"
      }
    ],
    "name": "NewLiquidationThreshold",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldMinLiquidatableCollateral",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newMinLiquidatableCollateral",
        "type": "uint256"
      }
    ],
    "name": "NewMinLiquidatableCollateral",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract ResilientOracleInterface",
        "name": "oldPriceOracle",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "contract ResilientOracleInterface",
        "name": "newPriceOracle",
        "type": "address"
      }
    ],
    "name": "NewPriceOracle",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "rewardsDistributor",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "rewardToken",
        "type": "address"
      }
    ],
    "name": "NewRewardsDistributor",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newSupplyCap",
        "type": "uint256"
      }
    ],
    "name": "NewSupplyCap",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferStarted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "acceptOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "accessControlManager",
    "outputs": [
      {
        "internalType": "contract IAccessControlManagerV8",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "accountAssets",
    "outputs": [
      {
        "internalType": "contract VToken",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "market",
        "type": "address"
      },
      {
        "internalType": "enum ComptrollerStorage.Action",
        "name": "action",
        "type": "uint8"
      }
    ],
    "name": "actionPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract RewardsDistributor",
        "name": "_rewardsDistributor",
        "type": "address"
      }
    ],
    "name": "addRewardsDistributor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "allMarkets",
    "outputs": [
      {
        "internalType": "contract VToken",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "borrowCaps",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "checkMembership",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "closeFactorMantissa",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "vTokens",
        "type": "address[]"
      }
    ],
    "name": "enterMarkets",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vTokenAddress",
        "type": "address"
      }
    ],
    "name": "exitMarket",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getAccountLiquidity",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "error",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "liquidity",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "shortfall",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAllMarkets",
    "outputs": [
      {
        "internalType": "contract VToken[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getAssetsIn",
    "outputs": [
      {
        "internalType": "contract VToken[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getBorrowingPower",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "error",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "liquidity",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "shortfall",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "vTokenModify",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "redeemTokens",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "borrowAmount",
        "type": "uint256"
      }
    ],
    "name": "getHypotheticalAccountLiquidity",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "error",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "liquidity",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "shortfall",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getRewardDistributors",
    "outputs": [
      {
        "internalType": "contract RewardsDistributor[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "getRewardsByMarket",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "rewardToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "supplySpeed",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowSpeed",
            "type": "uint256"
          }
        ],
        "internalType": "struct ComptrollerStorage.RewardSpeeds[]",
        "name": "rewardSpeeds",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "healAccount",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "loopLimit",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "accessControlManager",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isComptroller",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "isDeprecated",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "isMarketListed",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "contract VToken",
            "name": "vTokenCollateral",
            "type": "address"
          },
          {
            "internalType": "contract VToken",
            "name": "vTokenBorrowed",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "repayAmount",
            "type": "uint256"
          }
        ],
        "internalType": "struct ComptrollerStorage.LiquidationOrder[]",
        "name": "orders",
        "type": "tuple[]"
      }
    ],
    "name": "liquidateAccount",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vTokenBorrowed",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "vTokenCollateral",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "actualRepayAmount",
        "type": "uint256"
      }
    ],
    "name": "liquidateCalculateSeizeTokens",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "error",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "tokensToSeize",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "liquidationIncentiveMantissa",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "markets",
    "outputs": [
      {
        "internalType": "bool",
        "name": "isListed",
        "type": "bool"
      },
      {
        "internalType": "uint256",
        "name": "collateralFactorMantissa",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "liquidationThresholdMantissa",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "maxLoopsLimit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "minLiquidatableCollateral",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "oracle",
    "outputs": [
      {
        "internalType": "contract ResilientOracleInterface",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pendingOwner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "poolRegistry",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "borrowAmount",
        "type": "uint256"
      }
    ],
    "name": "preBorrowHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vTokenBorrowed",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "vTokenCollateral",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "repayAmount",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "skipLiquidityCheck",
        "type": "bool"
      }
    ],
    "name": "preLiquidateHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "minter",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "mintAmount",
        "type": "uint256"
      }
    ],
    "name": "preMintHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "redeemer",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "redeemTokens",
        "type": "uint256"
      }
    ],
    "name": "preRedeemHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      }
    ],
    "name": "preRepayHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vTokenCollateral",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "seizerContract",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "liquidator",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      }
    ],
    "name": "preSeizeHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "transferTokens",
        "type": "uint256"
      }
    ],
    "name": "preTransferHook",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "accessControlManager_",
        "type": "address"
      }
    ],
    "name": "setAccessControlManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken[]",
        "name": "marketsList",
        "type": "address[]"
      },
      {
        "internalType": "enum ComptrollerStorage.Action[]",
        "name": "actionsList",
        "type": "uint8[]"
      },
      {
        "internalType": "bool",
        "name": "paused",
        "type": "bool"
      }
    ],
    "name": "setActionsPaused",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "newCloseFactorMantissa",
        "type": "uint256"
      }
    ],
    "name": "setCloseFactor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "newCollateralFactorMantissa",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "newLiquidationThresholdMantissa",
        "type": "uint256"
      }
    ],
    "name": "setCollateralFactor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "newLiquidationIncentiveMantissa",
        "type": "uint256"
      }
    ],
    "name": "setLiquidationIncentive",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken[]",
        "name": "vTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "newBorrowCaps",
        "type": "uint256[]"
      }
    ],
    "name": "setMarketBorrowCaps",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken[]",
        "name": "vTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "newSupplyCaps",
        "type": "uint256[]"
      }
    ],
    "name": "setMarketSupplyCaps",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "limit",
        "type": "uint256"
      }
    ],
    "name": "setMaxLoopsLimit",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "newMinLiquidatableCollateral",
        "type": "uint256"
      }
    ],
    "name": "setMinLiquidatableCollateral",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ResilientOracleInterface",
        "name": "newOracle",
        "type": "address"
      }
    ],
    "name": "setPriceOracle",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "supplyCaps",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "supportMarket",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "updatePrices",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
""")
        